import { execFile } from "node:child_process";
import os from "node:os";
import process from "node:process";
import util from "node:util";
export class FfmpegCodecs {
    constructor(platform) {
        this._gpuMem = 0;
        this.log = platform.log;
        this.platform = platform;
        this.videoProcessor = platform.config.videoProcessor;
        this.videoProcessorCodecs = {};
        this.videoProcessorHwAccels = {};
    }
    // Launch our configured controllers once all accessories have been loaded. Once we do, they will sustain themselves.
    async probe() {
        // Let's conduct our system-specific capability probes.
        switch (this.platform.hostSystem) {
            case "raspbian":
                // If we're on a Raspberry Pi, let's verify that we have enough GPU memory for hardware-based decoding and encoding.
                await this.probeRpiGpuMem();
                break;
            default:
                break;
        }
        // First things first - ensure we've got a working video processor before we do anything else.
        if (!(await this.probeVideoProcessorCodecs()) || !(await this.probeVideoProcessorHwAccel())) {
            return false;
        }
        return true;
    }
    // Utility to determine whether or not a specific decoder is available to the video processor for a given format.
    hasDecoder(codec, decoder) {
        // Normalize our lookups.
        codec = codec.toLowerCase();
        decoder = decoder.toLowerCase();
        return this.videoProcessorCodecs[codec]?.decoders.some(x => x === decoder);
    }
    // Utility to determine whether or not a specific encoder is available to the video processor for a given format.
    hasEncoder(codec, encoder) {
        // Normalize our lookups.
        codec = codec.toLowerCase();
        encoder = encoder.toLowerCase();
        return this.videoProcessorCodecs[codec]?.encoders.some(x => x === encoder);
    }
    // Utility to determine whether or not a specific decoder is available to the video processor for a given format.
    hasHwAccel(accel) {
        return this.videoProcessorHwAccels[accel.toLowerCase()] ? true : false;
    }
    get gpuMem() {
        return this._gpuMem;
    }
    // Probe our video processor's hardware acceleration capabilities.
    async probeVideoProcessorHwAccel() {
        return this.probeCmd(this.videoProcessor, ["-hide_banner", "-hwaccels"], (stdout) => {
            // Iterate through each line, and a build a list of encoders.
            for (const accel of stdout.split(os.EOL)) {
                // Skip blank lines.
                if (!accel.length) {
                    continue;
                }
                // Skip the first line.
                if (accel === "Hardware acceleration methods:") {
                    continue;
                }
                // We've found a hardware acceleration method, let's add it.
                this.videoProcessorHwAccels[accel.toLowerCase()] = true;
            }
        });
    }
    // Probe our video processor's encoding and decoding capabilities.
    async probeVideoProcessorCodecs() {
        return this.probeCmd(this.videoProcessor, ["-hide_banner", "-codecs"], (stdout) => {
            // A regular expression to parse out the codec and it's supported decoders.
            const decodersRegex = /\S+\s+(\S+).+\(decoders: (.*?)\s*\)/;
            // A regular expression to parse out the codec and it's supported encoders.
            const encodersRegex = /\S+\s+(\S+).+\(encoders: (.*?)\s*\)/;
            // Iterate through each line, and a build a list of encoders.
            for (const codecLine of stdout.split(os.EOL)) {
                // Let's see if we have decoders.
                const decodersMatch = decodersRegex.exec(codecLine);
                // Let's see if we have encoders.
                const encodersMatch = encodersRegex.exec(codecLine);
                // If we found decoders, add them to our list of supported decoders for this format.
                if (decodersMatch) {
                    this.videoProcessorCodecs[decodersMatch[1]] = { decoders: [], encoders: [] };
                    this.videoProcessorCodecs[decodersMatch[1]].decoders = decodersMatch[2].split(" ").map(x => x.toLowerCase());
                }
                // If we found decoders, add them to our list of supported decoders for this format.
                if (encodersMatch) {
                    if (!this.videoProcessorCodecs[encodersMatch[1]]) {
                        this.videoProcessorCodecs[encodersMatch[1]] = { decoders: [], encoders: [] };
                    }
                    this.videoProcessorCodecs[encodersMatch[1]].encoders = encodersMatch[2].split(" ").map(x => x.toLowerCase());
                }
            }
        });
    }
    // Probe Raspberry Pi GPU.
    async probeRpiGpuMem() {
        return this.probeCmd("vcgencmd", ["get_mem", "gpu"], (stdout) => {
            // A regular expression to parse out the configured GPU memory on the Raspberry Pi.
            const gpuRegex = /^gpu=(.*)M\n$/;
            // Let's see what we've got.
            const gpuMatch = gpuRegex.exec(stdout);
            // We matched what we're looking for.
            if (gpuMatch) {
                // Parse the result and retrieve our allocated GPU memory.
                this._gpuMem = parseInt(gpuMatch[1]);
                // Something went wrong.
                if (isNaN(this._gpuMem)) {
                    this._gpuMem = 0;
                }
            }
        });
    }
    // Utility to probe the capabilities of FFmpeg and the host platform.
    async probeCmd(command, commandLineArgs, processOutput) {
        try {
            // Promisify exec to allow us to wait for it asynchronously.
            const execAsync = util.promisify(execFile);
            // Check for the codecs in our video processor.
            const { stdout } = await execAsync(command, commandLineArgs);
            processOutput(stdout);
            return true;
        }
        catch (error) {
            // It's really a SystemError, but Node hides that type from us for esoteric reasons.
            if (error instanceof Error) {
                const execError = error;
                if (execError.code === "ENOENT") {
                    this.log.error("Unable to find '%s' in path: '%s'.", command, process.env.PATH);
                }
                else {
                    this.log.error("Error running %s: %s", command, error.message);
                }
            }
            this.log.error("Unable to probe the capabilities of your Homebridge host without access to '%s'. Ensure that it is available in your path and correctly working.", command);
            return false;
        }
    }
}
//# sourceMappingURL=protect-ffmpeg-codecs.js.map